use std::{
    env::{self, current_dir},
    path::PathBuf,
    sync::{Arc, Mutex},
};

use anyhow::anyhow;
use clap::Command;
use git2::Repository;

use crate::{
    actions,
    context::{CuddleContext, CuddleTreeType},
    model::{CuddleScript, CuddleShellScriptArg},
};

#[derive(Debug, Clone)]
#[allow(dead_code)]
struct CuddleAction {
    script: CuddleScript,
    path: PathBuf,
    name: String,
}
#[allow(dead_code)]
impl CuddleAction {
    pub fn new(script: CuddleScript, path: PathBuf, name: String) -> Self {
        Self { script, path, name }
    }

    pub fn execute(self, variables: Vec<CuddleVariable>) -> anyhow::Result<()> {
        match self.script {
            CuddleScript::Shell(s) => {
                let mut arg_variables: Vec<CuddleVariable> = vec![];
                if let Some(args) = s.args {
                    for (k, v) in args {
                        let var = match v {
                            CuddleShellScriptArg::Env(e) => {
                                let env_var = match env::var(e.key.clone()) {
                                    Ok(var) => var,
                                    Err(e) => {
                                        log::error!("env_variable not found: {}", k);
                                        return Err(anyhow::anyhow!(e));
                                    }
                                };
                                CuddleVariable::new(k.clone(), env_var)
                            }
                        };

                        arg_variables.push(var);
                    }
                } else {
                    arg_variables = vec![]
                };

                let mut vars = variables.clone();
                vars.append(&mut arg_variables);

                log::trace!("preparing to run action");

                match actions::shell::ShellAction::new(
                    self.name.clone(),
                    format!(
                        "{}/scripts/{}.sh",
                        self.path
                            .to_str()
                            .expect("action doesn't have a name, this should never happen"),
                        self.name
                    ),
                )
                .execute(vars)
                {
                    Ok(()) => {
                        log::trace!("finished running action");
                        Ok(())
                    }
                    Err(e) => {
                        log::error!("{}", e);
                        Err(e)
                    }
                }
            }
            CuddleScript::Dagger(_d) => Err(anyhow::anyhow!("not implemented yet!")),
        }
    }
}

#[derive(Debug, Clone)]
#[allow(dead_code)]
pub struct CuddleVariable {
    pub name: String,
    pub value: String,
}

impl CuddleVariable {
    pub fn new(name: String, value: String) -> Self {
        Self { name, value }
    }
}

#[derive(Debug)]
struct GitCommit {
    commit_sha: String,
}

impl GitCommit {
    fn new() -> anyhow::Result<GitCommit> {
        let repo = Repository::open(current_dir().expect("having current_dir available")).map_err(
            |e| {
                log::debug!("{}", e);
                anyhow::anyhow!("could not open repository")
            },
        )?;
        let head_ref = repo
            .head()
            .map_err(|e| {
                log::warn!("{}", e);
                anyhow::anyhow!("could not get HEAD")
            })?
            .target()
            .ok_or(anyhow::anyhow!(
                "could not extract head -> target to commit_sha"
            ))?;

        Ok(Self {
            commit_sha: head_ref.to_string(),
        })
    }
}

#[derive(Debug, Clone)]
pub struct CuddleCli<'a> {
    scripts: Vec<CuddleAction>,
    variables: Vec<CuddleVariable>,
    context: Arc<Mutex<Vec<CuddleContext>>>,
    command: Option<Command<'a>>,
    tmp_dir: Option<PathBuf>,
}

impl<'a> CuddleCli<'a> {
    pub fn new(context: Arc<Mutex<Vec<CuddleContext>>>) -> anyhow::Result<CuddleCli<'a>> {
        let mut cli = CuddleCli {
            scripts: vec![],
            variables: vec![],
            context: context.clone(),
            command: None,
            tmp_dir: None,
        };

        cli = cli
            .process_variables()
            .process_scripts()
            .process_templates()?
            .build_cli();

        Ok(cli)
    }

    fn process_variables(mut self) -> Self {
        if let Ok(context_iter) = self.context.clone().lock() {
            for ctx in context_iter.iter() {
                if let Some(variables) = ctx.plan.vars.clone() {
                    for (name, var) in variables {
                        self.variables.push(CuddleVariable::new(name, var))
                    }
                }

                if let CuddleTreeType::Root = ctx.node_type {
                    let mut temp_path = ctx.path.clone();
                    temp_path.push(".cuddle/tmp/");

                    self.variables.push(CuddleVariable::new(
                        "tmp".into(),
                        temp_path.clone().to_string_lossy().to_string(),
                    ));

                    self.tmp_dir = Some(temp_path);
                }
            }
        }

        match GitCommit::new() {
            Ok(commit) => self.variables.push(CuddleVariable::new(
                "commit_sha".into(),
                commit.commit_sha.clone(),
            )),
            Err(e) => {
                log::debug!("{}", e);
            }
        }

        self
    }

    fn process_scripts(mut self) -> Self {
        if let Ok(context_iter) = self.context.clone().lock() {
            for ctx in context_iter.iter() {
                if let Some(scripts) = ctx.plan.scripts.clone() {
                    for (name, script) in scripts {
                        self.scripts
                            .push(CuddleAction::new(script.clone(), ctx.path.clone(), name))
                    }
                }
            }
        }

        self
    }

    fn process_templates(self) -> anyhow::Result<Self> {
        if let None = self.tmp_dir {
            log::debug!("cannot process template as bare bones cli");
            return Ok(self);
        }

        // Make sure tmp_dir exists and clean it up first
        let tmp_dir = self
            .tmp_dir
            .clone()
            .ok_or(anyhow::anyhow!("tmp_dir does not exist aborting"))?;
        if tmp_dir.exists() && tmp_dir.ends_with("tmp") {
            std::fs::remove_dir_all(tmp_dir.clone())?;
        }
        std::fs::create_dir_all(tmp_dir.clone())?;
        // Handle all templating with variables and such.
        // TODO: use actual templating engine, for new we just copy templates to the final folder

        if let Ok(context_iter) = self.context.clone().lock() {
            for ctx in context_iter.iter() {
                let mut template_path = ctx.path.clone();
                template_path.push("templates");

                log::trace!("template path: {}", template_path.clone().to_string_lossy());

                if !template_path.exists() {
                    continue;
                }

                for file in std::fs::read_dir(template_path)?.into_iter() {
                    let f = file?;
                    let mut dest_file = tmp_dir.clone();
                    dest_file.push(f.file_name());

                    std::fs::copy(f.path(), dest_file)?;
                }
            }
        }

        Ok(self)
    }

    fn build_cli(mut self) -> Self {
        let mut root_cmd = Command::new("cuddle")
            .version("1.0")
            .author("kjuulh <contact@kasperhermansen.com>")
            .about("cuddle is your domain specific organization tool. It enabled widespread sharing through repositories, as well as collaborating while maintaining speed and integrity")
            .subcommand_required(true)
            .propagate_version(true);

        if self.scripts.len() > 0 {
            let mut execute_cmd = Command::new("x").about("x is your entry into your domains scripts, scripts inherited from parents will also be present here").subcommand_required(true);

            for script in self.scripts.iter() {
                let action_cmd = Command::new(script.name.clone());

                // TODO: Some way to add an about for clap, requires conversion from String -> &str

                execute_cmd = execute_cmd.subcommand(action_cmd);
            }

            root_cmd = root_cmd.subcommand(execute_cmd);
        }

        self.command = Some(root_cmd);

        self
    }

    pub fn execute(self) -> anyhow::Result<Self> {
        if let Some(mut cli) = self.command.clone() {
            let matches = cli.clone().get_matches();

            let res = match matches.subcommand() {
                Some(("x", exe_submatch)) => {
                    log::trace!("executing x");

                    match exe_submatch.subcommand() {
                        Some((name, _action_matches)) => {
                            log::trace!(action=name; "running action; name={}", name);
                            match self.scripts.iter().find(|ele| ele.name == name) {
                                Some(script) => {
                                    script.clone().execute(self.variables.clone())?;
                                    Ok(())
                                }
                                _ => Err(anyhow!("could not find a match")),
                            }
                        }
                        _ => Err(anyhow!("could not find a match")),
                    }
                }
                _ => Err(anyhow!("could not find a match")),
            };

            match res {
                Ok(()) => {}
                Err(e) => {
                    let _ = cli.print_long_help();
                    return Err(e);
                }
            }
        }

        Ok(self)
    }
}
